/* Asteroids3D - a first person game of blowing up asteroids
 * Copyright (C) 2000 Stuart Mark Pomerantz
 *
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  Stuart Pomerantz
 *  3935 Stonecliffe drive
 *  Monroeville, PA 15146
 *  stuart@pitt.edu
 */

#include "asteroids3D.h"


/**
 ** These variables are ALL the global variables in the game 
 **/


/* Flags for user preferences 0=off 1=on */
int axes_flag = 0 ;                /* turn on and off the absolute x,y,z axes. I use this for debugging. */
int crosshair_flag = 1 ;           /* turn the blue crosshair on or off */
int draw_tracer_flag = 1 ;         /* should the torpedo tracer light be drawn */ 
int draw_shield_flag = 1 ;         /* should the green alpha-blended shield effect be used */
int draw_dust_flag = 1 ;           /* should the dust particles be drawn */ 
int texture_flag = 0 ;             /* should the asteroids be textured */
int nebula_texture_flag = 0 ;      /* should the nebula textures be used */
int play_sound_flag = 1 ;          /* should sound be used */
int torp_texture_flag = 0 ;        /* should the torps be textured */
int draw_torp_lensflare_flag = 0 ; /* should the torp lens flares be drawn */

/* more user preferences */
float mouse_xdir = 1 ;             /* 1 = mouse movement turns the normal direction, -1 = mouse movement turns opposite direction (should be -1 or 1 only) */
float mouse_ydir = 1 ;             /* 1 = mouse movement turns the normal direction, -1 = mouse movement turns opposite direction (should be -1 or 1 only) */
int num_asteroids = NUM_ASTEROIDS ;/* number of asteroids, set by #define or command line */
int windowx = WINDOWX ; /* window x size, changed in reshape(). set by #define or on command line */
int windowy = WINDOWY ; /* window y size, changed in reshape(). set by #define or on command line */


/* Flags for internal program control */
int torps_in_flight_flag = 0 ;   /* are any torpedos in flight */
int mouse_roty_flag = 0 ;        /* set to 1 if the mouse has been moved out of the dead zone to rotate the ship about the y axis */
int mouse_rotx_flag = 0 ;        /* set to 1 if the mouse has been moved out of the dead zone to rotate the ship about the x axis */
int shield_activated_flag = 0 ;  /* has the shield been activated */
int explosions_active_flag = 0 ; /* are there explosions activated */
int paused_flag = 0 ;            /* is the game paused */
int game_end_flag = 0 ;          /* is the game over */
int tracer_light_flag = 0 ;      /* turn on the tracer light */



/* eye's local coordinate system. */
vector xaxis , yaxis , zaxis ;

/* eye's position in space */
vector pos ;

/* light0 ( the sun ) position,  
 * light1 (headlight) position & direction 
 * light2 torpedo tracer
 */
float l0_pos[4], l1_pos[4], l1_dir[4], l2_pos[4] ;

/* three important vectors: 
 *
 * 1) thrust 
 *
 * is the the vector which sums the thrust vectors
 * which the user has input but which have not yet been applied 
 * this is a "request"
 *
 * 2) velocity
 *
 * is the sum of the parts of the thrust vector which have been
 * applied (by the engine) and are now acting 
 * ( in a newtonian sense ) to move the ship.
 *
 * 3) rotation
 *
 * is the vector which holds the input requests for a rotation about an axis.
 * this is a "request" modulated by dt
 *
 */
vector thrust, velocity, rotation ;


/* define a torpedo array to hold, say, MAX_TORPS worth of them */
torpedo torp[MAX_TORPS] ;
torpedo sorted_torp[MAX_TORPS] ; /* doing this so alpha blending works correctly, this is scratch space */
int torp_dl ; /* torpedo display list */
int torp_tex ; /* torpedo billboard texture */
int torp_billboard_dl ; /* billboard for torpedo texture */

/* torpedo trails */
torpedo_trail torp_trail[MAX_TORP_TRAILS] ;

/* torp gauge dl 0 = off, 1 = on */
int torp_gauge_dl[2] ;

/* asteroid linked list head and tail pointers */
node *al_head, al_tail ;

/* shield display list, shield ttl, and shield strength */
int shield_dl ;
int shield_ttl ;
float shield_strength ;

/* crosshair display list */
int crossh_dl ;

/* sun display list */
int sun_dl ;

/* asteroid display list */
int ast_dl[NUM_AST_TYPES] ;

/* asteroid texture object list */
int ast_tex[NUM_AST_TYPES] ;

/* list of asteroids to be deleted each frame.
 * This list actually contains the node PRIOR to the node
 * to be deleted ( so I can hook PRIOR node to the deleted node's
 * "next" node. Hence the name. In any given frame, the largest
 * number of asteroids that can possibly be deleted is MAX_TORPS
 * so I make an array of this number of node pointers, just in case.
 * initialized in init_asteroids()
 */
node *delete_next_ast[MAX_TORPS] ;


/* explosion array, which is defined to be MAX_BLASTS > MAX_TORPS size since
 * there may be more than MAX_TORPS number of simultanious explosions
 * since the explosions (currently) last longer then the torp ttl.
 */
blast explosion[MAX_BLASTS] ;
int   expl_dl[NUM_BLAST_TYPES];
int   expl_winshield_particle_dl ;
int   expl_winshield_particle_count = 0 ;

/* track the mouse cursor */
int mouse_x, mouse_y ;

/* the proportion the mouse has moved out of the dead zone */
float mouse_xprop = 0 ;
float mouse_yprop = 0 ;

/* dust particle array */
vector dust_part[NUM_DUST_PART][NUM_CHILD_DUST_PART] ;

/* target box display list */
int target_box_dl, target_circle_dl ;
int target_arrow_on_dl, target_arrow_off_dl ;
int target_steering_cue_dl ;

/* a pointer to, well, that's pretty obvious */
node *currently_targeted_asteroid = NULL ;

/* nebula billboard texture object list, and nebula display list */
int nebula_tex[NUM_NEBULA_TEX] ;
int nebula_dl[NUM_NEBULA_TEX] ;

/* number of asteroids (for ship's "scanner" ) */
int num_asteroids_on_scanner ;


/* for collision warning gauge 0 = off, 1 = on */
int cwg_rear_dl[2] ; 
int cwg_above_dl[2], cwg_below_dl[2] ;
int cwg_left_dl[2],  cwg_right_dl[2] ;
int cwg_rear, cwg_above, cwg_below, cwg_left, cwg_right ; /* also 0 = off, 1 = on  */

/* windsheid particle array */
wind_part winshield_particle[MAX_BLASTS*NUM_WP_PER_BLAST] ;


/* last but not least, what would a game be without a score */
unsigned int score = 0 ;


/* time to end hud gauge flicker, intially off. Turned on after a collision. */
float flicker_hud_ttl = 0 ;


/* the number of game time ticks that have past */
unsigned long ticks = 0 ;

/* dt = delta time, the amount of time (in seconds)
 *  which passes with each game tick 
 */
float dt = 1.0/TPS ;

/* this is the number of ticks per second, which I may want to change dynamically */
float tps = TPS ;


/* frame calculation thread id, and the run_yield
 * semaphore for interleaving calculation and display updates
 */
pthread_t fc_tid ;
sem_t run_yield ;

 
/* OSS/Free audio device file descriptor */
int audio_device_fd ;

/* asteroid sounds */
sound ast_snd[NUM_AST_TYPES] ;

/* my sound buffer (doing this so sounds can be interrupted) */
char sound_buffer[MAX_SOUND_BUF_BYTES] ;
int  sound_buffer_start = 0, sound_buffer_end = 0 ;

/* target aquired sound */
sound ta_snd ;


/* data structure to hold the scores read from the score file */
score_file_entry score_entry[NUM_SCORES] ;

/* lens flares */
lens_flare sun_flare[MAX_FLARES] ;
lens_flare torp_flare[MAX_FLARES] ;







int main(int argc, char **argv) {

	display_copyright_on_terminal() ;

	parse_command_line( argc, argv ) ;
	
	glut_initialization(&argc, argv) ;
	gl_initialization() ;
	my_initialization() ;

	glutMainLoop();
	
	return(0) ;
}


void glut_initialization(int *argc, char **argv) {

	/* pass command line args to glut */
	glutInit(argc, argv);
	
	glutInitDisplayMode(GLUT_DOUBLE|GLUT_RGB|GLUT_DEPTH);
	
	glutInitWindowSize(windowx, windowy);
	glutInitWindowPosition(150, 50);
	glutCreateWindow(COPYRIGHT);
	

	glutDisplayFunc( display );
	glutMouseFunc( mouse_handler ); 
	glutKeyboardFunc( keyboard_handler );
	glutIdleFunc( idle ) ;
	glutReshapeFunc( reshape ) ;
	glutPassiveMotionFunc( passive_mouse_handler ) ;
	

	/* glutTimerFunc( MSEC, check_collision, 1) ;	 */

	/* center the mouse cursor -- see glutPassiveMouseFunc */
	glutWarpPointer(windowx/2, windowy/2) ; 
	
	/* looks better than the arrow, I think */
	/* glutSetCursor(GLUT_CURSOR_CROSSHAIR) ;*/
	glutSetCursor(GLUT_CURSOR_NONE) ;
	
	return ;
}

void gl_initialization(void) {
		
	glClearColor(0.0, 0.0, 0.0, 0.0);
	 
	glMatrixMode(GL_PROJECTION); 
   	glLoadIdentity();
	glFrustum( LEFT, RIGHT, BOTTOM, TOP, ZNEAR, ZFAR ) ;
	 
	glEnable(GL_CULL_FACE) ;
	
	glEnable(GL_DEPTH_TEST);

	glEnable(GL_TEXTURE_2D) ;
		 
	return ;
}


void my_initialization(void) {

	printf("Initialising...\n");

	reset_view() ;
	gl_init_light() ;	
	gl_init_headlight() ; /* reset_view() must be called before gl_init_headlight */
	init_thrust() ;
	init_torps() ;
	init_textures();
	init_asteroids() ;
	init_shield()	 ;
	init_hud() ;
	init_sun() ; /* must be called after  gl_init_light */
	init_explosions();	
	init_dust() ;
	init_target_dls() ;
	init_nebula_billboards() ; /* must be called after  gl_init_light, and init_textures */
	init_wind_parts() ;
	init_sound() ;
	init_lens_flares() ;


	/* finally, and always last,
	 * set up the run_yield semaphore and
	 * crank up the frame calculation thread 
	 */
	
	sem_init( &run_yield, 0 , 1) ;

	pthread_create( &fc_tid, NULL, (void *)calculate_frame, NULL );
	
	printf("Done.\n");	
	
	return ;
}


void idle(void) {

	/* calculate_frame() is now a thread by itself, hope this works */
	/* calculate_frame() ;*/
		
	glutPostRedisplay() ;
	
	return ;
}


void reshape(int w, int h) {
	
	glViewport(0, 0, (GLsizei) w, (GLsizei) h);

	windowx = w ;
	windowy = h ;
	
	return ;
}



void keyboard_handler( unsigned char key, int x, int y ) {
	
	
	switch( key ) {
	
		case 27:       /* escape key exits */
			end_game() ;
			break ;			


		/* activate the debugging axes */			
			
		case 'a':
			axes_flag = 1 ;
			break ;
			
		case 'A':
			axes_flag = 0 ;
			break ;

		/* 
		 * driving keys  -- 'f' is the keyboard centerpoint
		 *                  and using the keypad
		 */

		case 'r':
			slide_alongz(MAX_TINPUT) ;
			break ;

		case 'v':
			slide_alongz(-MAX_TINPUT) ;
			break ;

		case 'd':
			slide_alongx(MAX_TINPUT) ;
			break ;

		case 'g':
			slide_alongx(-MAX_TINPUT) ;
			break ;

		case 'e':
			slide_alongy(MAX_TINPUT) ;
			break ;

		case 't':
			slide_alongy(-MAX_TINPUT) ;
			break ;
		
		case 'f':
			stop() ;
			break ;
				
		case '4':
			turn_abouty(-MAX_DPS_INPUT) ;
			break; 

		case '6':
			turn_abouty(MAX_DPS_INPUT) ;
			break; 

		case '7':
			turn_aboutz(MAX_DPS_INPUT) ;
			break ;
			
		case '9':
			turn_aboutz(-MAX_DPS_INPUT) ;
			break ;			

		case '8':
			turn_aboutx(-MAX_DPS_INPUT) ;
			break ;
			
		case '2':
			turn_aboutx(MAX_DPS_INPUT) ;
			break ;

		/*
		 * features
		 */
		
		case '[':
			activate_headlight() ;
			break ;
			
		case '{':
			deactivate_headlight() ;
			break ;

		case '+':
			crosshair_flag = 1 ;
			break ;

		case '=':
			crosshair_flag = 0 ;
			break ;
			
		case ']':
			draw_tracer_flag = 1 ;
			break ;
			
		case '}':
			draw_tracer_flag = 0 ;
			break ;
			
		case '-':
			draw_shield_flag = 1;
			break ;
			
		case '_':
			draw_shield_flag = 0 ;
			break ;
			
		case 'z':
			draw_dust_flag = 1 ;
			break ;
		
		case 'Z':
			draw_dust_flag = 0 ;
			break ;
			
		case 'i':
			texture_flag = 1 ;
			break; 

		case 'I':
			texture_flag = 0 ;
			break; 

		case 'n':
			nebula_texture_flag = 1 ;
			break ;

		case 'N':
			nebula_texture_flag = 0 ;
			break ;
			
		case 'b':
			target_asteroid() ;
			break ;

		case '\t':
			auto_target_asteroid() ;
			break ;

		case 'p':
			start_pause() ;
			break ;

		case '.':
			play_sound_flag = 1 ;
			break ;

		case '>':
			play_sound_flag = 0 ;
			break ;
			
		case 'm':
			torp_texture_flag = 1 ;
			break ;

		case 'M':
			torp_texture_flag = 0 ;
			break ;
			
		case '1':
			mouse_xdir = -1 ;
			break ;

		case '!':
			mouse_xdir = 1 ;
			break ;

		case '3':
			mouse_ydir = -1 ;
			break ;

		case '#':
			mouse_ydir = 1 ;
			break ;			

		case '5':
			draw_torp_lensflare_flag = 1 ;
			break ;

		case '%':
			draw_torp_lensflare_flag = 0 ;
			break ;

		/* finally, the ever important FIRE button */
		
		case ' ':
			fire_torpedo() ;
			break ;
				
		default:
			printf("got key %c, x=%d,y=%d\n",key,x,y) ;	
			break ;
	}
	
	return ;
}




void mouse_handler( int button, int state, int x, int y) {

	/* mouse x,y are given in window coordinates
	 */


	switch( button ) {
	
		case GLUT_LEFT_BUTTON:
			if( state == GLUT_DOWN ) {
				fire_torpedo() ;
			}
		break ;


		case GLUT_RIGHT_BUTTON:
			if( state == GLUT_DOWN ) {
			}
		break ;
	}
	
	return ;
}

void passive_mouse_handler(int x, int y) {
	int left_bound   = windowx/2 - 50 ; /* dead zone boundaries */
	int right_bound  = windowx/2 + 50 ;
	int top_bound    = windowy/2 - 50 ;
	int bottom_bound = windowy/2 + 50 ;

	/* set the mouse position globals */
	mouse_x = x ;
	mouse_y = y ;

	
	/* if mouse is inside dead zone */
	if(  (x > left_bound) && (x < right_bound) &&  
		 (y > top_bound ) && (y < bottom_bound) ) {
		mouse_roty_flag = 0 ;
		mouse_rotx_flag = 0 ;
		return ;
	}

	
	/* if mouse is not in the dead zone set the flag */
	
	if( x > right_bound) {
		mouse_roty_flag = 1 ;
		mouse_yprop = (float)(x-right_bound)/(float)(windowx-right_bound) ;
	}
		

	if( x < left_bound) {
		mouse_roty_flag = 1 ;
		mouse_yprop = -1*(1 - (float)(x)/(float)left_bound) ;
	}

	
	if( y < top_bound) {
		mouse_rotx_flag = 1 ;
		mouse_xprop = (1 - (float)(y)/(float)top_bound) ;
	}


	if( y > bottom_bound) {
		mouse_rotx_flag = 1 ;	
		mouse_xprop = -1*(float)(y-bottom_bound)/(float)(windowy-bottom_bound)  ;		
	}

	/* for reversing mouse direction */
	
	mouse_yprop = mouse_yprop*mouse_ydir ;

	mouse_xprop = mouse_xprop*mouse_xdir ;
	
	return ;
}

void end_game(void) {

	game_end_flag = 1 ;  /* this causes the calculation thread to exit */
	init_score_display() ;

	return ;
}


void parse_command_line(int argc, char **argv) {
	int i ;
	int temp1, temp2 ;

	for( i = 0 ; i < argc ; i++ ) {

		if( strcmp("-help",argv[i]) == 0 ) {
			printf("valid command line options:\n") ;
			printf("   -num <number>\n"
			       "         to set the number of asteroids\n\n") ;
			printf("   -geom <width>x<height>\n"
			       "         to set the geometry of the window (in pixels)\n\n") ;
			exit(0) ;
		}

		if( strcmp("-num",argv[i]) == 0 ) {
			if( i + 1 < argc ) {
				if( sscanf(argv[i+1],"%d\n",&temp1) == 1 ) {
					if( temp1 > 1 )
						num_asteroids = temp1 ;
					i++ ;
				}
			}
			continue ;
		}


		if( strcmp("-geom",argv[i]) == 0 ) {
			if( i + 1 < argc ) {
				if( sscanf(argv[i+1],"%dx%d\n",&temp1,&temp2) == 2 ) {			
					if( (temp1 > 0) && (temp2 >0) ) {
						windowx = temp1 ;
						windowy = temp2 ;
					}
				}
			}
		}


	} /* end for each argument */

	return ;
}

void display_copyright_on_terminal(void) {
	printf(COPYRIGHT) ;
	printf("Asteroids3D comes with ABSOLUTELY NO WARRANTY; for details\n"
           "see the file COPYRIGHT which should have come with this program.\n"
		   "This is free software, and you are welcome to redistribute it\n"
		   "under certain conditions; see the file COPYRIGHT\n"
		   "which should have come with this program for details.\n") ;
}
